# SPDX-License-Identifier: BSD-2-Clause
import logging
import subprocess
from dataclasses import dataclass
from typing import Iterable


log = logging.getLogger(__name__)


@dataclass
class Dataset:
    name: str


@dataclass
class Snapshot:
    dataset: Dataset
    name: str


class ZfsOperationFailed(Exception):
    pass


def run(cmd: str) -> Iterable[str]:
    log.debug(f"Run: {cmd}")
    p = subprocess.run(args=cmd, shell=True, text=True, capture_output=True)
    if p.returncode != 0:
        raise ZfsOperationFailed(p.stderr.split("\n")[0])
    return p.stdout.rstrip("\n").split("\n")


def get_datasets():
    lines = run("zfs list -H -t filesystem -o name")
    return [Dataset(name=line) for line in lines]


def get_snapshots(dataset: Dataset):
    lines = run(f"zfs list -H -r -t snapshot -o name -S creation {dataset.name}")

    snapshots = []
    for line in lines:
        _, name = line.split("@")
        snapshots.append(Snapshot(dataset=dataset, name=name))
    return snapshots


def create_snapshot(dataset: Dataset, name: str):
    run(f"zfs snapshot {dataset.name}@{name}")


def destroy_snapshot(snapshot: Snapshot):
    run(f"zfs destroy {snapshot.dataset.name}@{snapshot.name}")



import logging
import sys
from pathlib import Path
from typing import Optional


def init_logging(
    log_level: str, log_file: Optional[Path] = None, log_console: bool = False
) -> None:
    log_level = getattr(logging, log_level)

    root_logger = logging.getLogger()
    root_logger.setLevel(log_level)

    if log_file:
        file_formatter = logging.Formatter(
            style="{", fmt="{asctime} {name:12} {levelname:8} {message}"
        )
        file_handler = logging.FileHandler(str(log_file))
        file_handler.setFormatter(file_formatter)
        file_handler.setLevel(logging.DEBUG)
        root_logger.addHandler(file_handler)

    if log_console:
        console_formatter = logging.Formatter(style="{", fmt="{message}")
        console_handler = logging.StreamHandler(stream=sys.stdout)
        console_handler.setFormatter(console_formatter)
        console_handler.setLevel(logging.DEBUG)
        root_logger.addHandler(console_handler)

    logging.getLogger("PidFile").setLevel(logging.ERROR)



# SPDX-License-Identifier: BSD-2-Clause
import datetime
import logging
import sys
from pathlib import Path
from typing import Optional

import click
import pid
import pkg_resources

from . import config, manage
from .logging import init_logging


log = logging.getLogger(__name__)


def load_config(config_file: Optional[Path] = None) -> Optional[dict]:
    if config_file is None:
        config_file = config.get_path()

    log.info(f"Loading config file from: {config_file}")

    try:
        with open(str(config_file), "r") as f:
            return config.load(data=f)

    except FileNotFoundError:
        log.error(f"Config file not found: {config_file}")

    except config.ValidationError as e:
        log.error(f"Config file invalid: {e}")


@click.group()
@click.pass_context
def program(ctx):
    ctx.obj["now"] = datetime.datetime.now()


@program.command("cron")
@click.option(
    "--config-file", type=click.Path(), default=None, help="Path to config file."
)
@click.option(
    "--log-file",
    type=click.Path(),
    default="/var/log/zsm.log",
    help="Path to log file.",
)
@click.option(
    "--log-level", type=str, default="INFO", help="Minimum log level to output."
)
@click.option("--log-console", is_flag=True, help="Write logs to console.")
@click.option(
    "--dry-run",
    is_flag=True,
    help="Prevent all operations that perform modifications from running.",
)
@click.pass_context
def cron(
    ctx,
    config_file: str,
    log_file: str,
    log_level: str,
    log_console: bool,
    dry_run: bool,
):
    if config_file is not None:
        config_file = Path(config_file)

    if log_file is not None:
        log_file = Path(log_file)

    init_logging(log_level=log_level, log_file=log_file, log_console=log_console)

    config_data = load_config(config_file=config_file)

    if config_data is None:
        sys.exit(1)

    try:
        with pid.PidFile("zsm_cron"):
            manage.manage_snapshots(
                config=config_data, now=ctx.obj["now"], dry_run=dry_run
            )

    except pid.PidFileAlreadyLockedError:
        log.info("Found existing pidfile, aborting.")
        sys.exit(2)


@program.command("validate-config")
@click.option(
    "--config-file", type=click.Path(), default=None, help="Path to config file."
)
@click.option(
    "--log-level", type=str, default="INFO", help="Minimum log level to output."
)
def validate_config(config_file: str, log_level: str):
    if config_file is not None:
        config_file = Path(config_file)

    init_logging(log_level=log_level, log_console=True)

    config_data = load_config(config_file=config_file)

    if config_data is None:
        sys.exit(1)

    log.info("Config file is valid!")


@program.command("version")
def version():
    version = pkg_resources.get_distribution("zsm").version
    print(f"zsm {version}")


def cli():
    program(obj={})



# SPDX-License-Identifier: BSD-2-Clause
import textwrap
from pathlib import Path
from unittest import mock

import marshmallow
import pytest

import zsm.config
import zsm.zfs


datasets = [
    zsm.zfs.Dataset(name="tank/a"),
    zsm.zfs.Dataset(name="tank/b"),
    zsm.zfs.Dataset(name="tank/c"),
]


@mock.patch("platform.system", mock.MagicMock(return_value="FreeBSD"))
def test_get_platform_path_returns_correct_path_on_freebsd():
    path = zsm.config.get_platform_path()

    assert path == Path("/usr/local/etc")


@mock.patch("platform.system", mock.MagicMock(return_value="Linux"))
def test_get_platform_path_returns_correct_path_on_linux():
    path = zsm.config.get_platform_path()

    assert path == Path("/etc")


def test_get_path():
    path = zsm.config.get_path()
    assert path == zsm.config.get_platform_path() / "zsm.yaml"


@mock.patch("zsm.zfs.get_datasets", mock.MagicMock(return_value=datasets))
def test_schema_raises_validationerror_for_invalid_dataset_name():
    data = {"snapshots": [{"dataset": "invalid/invalid"}]}

    with pytest.raises(marshmallow.ValidationError) as e:
        data = zsm.config.ConfigSchema().load(data)

    assert "Dataset does not exist" in e.value.messages["snapshots"][0]["dataset"]


@mock.patch("zsm.zfs.get_datasets", mock.MagicMock(return_value=datasets))
def test_load_works_with_valid_data():
    data = """\
    snapshots:
      - dataset: "tank/a"
        label: "hourly"
        frequency: "1h"
        retention: 24
    """

    zsm.config.load(data=textwrap.dedent(data))


@mock.patch("zsm.zfs.get_datasets", mock.MagicMock(return_value=datasets))
def test_load_raises_validationerror_for_invalid_datas():
    datas = [
        # Invalid YAML.
        "!@#",
        # Missing fields.
        """\
        snapshots:
          - dataset: "tank/a"
        """,
        # Label contains invalid character.
        """\
        snapshots:
          - dataset: "tank/a"
            label: "_"
            frequency: "1h"
            retention: 1
        """,
        # Frequency contains invalid unit.
        """\
        snapshots:
          - dataset: "tank/a"
            label: "a"
            frequency: "1p"
            retention: 1
        """,
    ]

    for data in datas:
        with pytest.raises(zsm.config.ValidationError):
            zsm.config.load(data=textwrap.dedent(data))



# SPDX-License-Identifier: BSD-2-Clause
from unittest import mock

import pytest

import zsm.zfs


def get_mocked_subprocess_run_pipeline(returncode=0, stdout="", stderr=""):
    def get_out_err(value):
        if isinstance(value, list):
            return "\n".join(value)
        return value

    mocked_pipeline = mock.MagicMock()
    mocked_pipeline.returncode = returncode
    mocked_pipeline.stdout = get_out_err(stdout)
    mocked_pipeline.stderr = get_out_err(stderr)

    return mocked_pipeline


@mock.patch(
    "subprocess.run",
    return_value=get_mocked_subprocess_run_pipeline(
        returncode=0,
    ),
)
def test_run_with_success(mocked_subprocess_run):
    zsm.zfs.run("asdf")
    mocked_subprocess_run.assert_called()


@mock.patch(
    "subprocess.run",
    return_value=get_mocked_subprocess_run_pipeline(returncode=1, stderr="a problem?"),
)
def test_run_with_failure(mocked_subprocess_run):
    with pytest.raises(zsm.zfs.ZfsOperationFailed) as e:
        zsm.zfs.run("asdf")

    mocked_subprocess_run.assert_called()
    assert str(e.value) == "a problem?"


@mock.patch(
    "subprocess.run",
    return_value=get_mocked_subprocess_run_pipeline(
        stdout=["tank/a", "tank/b"],
    ),
)
def test_get_datasets(mocked_subprocess_run):
    datasets = zsm.zfs.get_datasets()

    mocked_subprocess_run.assert_called()
    assert datasets[0].name == "tank/a"
    assert datasets[1].name == "tank/b"


@mock.patch(
    "subprocess.run",
    return_value=get_mocked_subprocess_run_pipeline(
        stdout=["tank/a@abc", "tank/a@xyz"],
    ),
)
def test_get_snapshots(mocked_subprocess_run):
    snapshots = zsm.zfs.get_snapshots(dataset=zsm.zfs.Dataset(name="tank/a"))

    mocked_subprocess_run.assert_called()
    assert snapshots[0].name == "abc"
    assert snapshots[1].name == "xyz"


@mock.patch(
    "subprocess.run",
    return_value=get_mocked_subprocess_run_pipeline(),
)
def test_create_snapshot(mocked_subprocess_run):
    zsm.zfs.create_snapshot(dataset=zsm.zfs.Dataset(name="tank/a"), name="asdf")
    mocked_subprocess_run.assert_called()


@mock.patch(
    "subprocess.run",
    return_value=get_mocked_subprocess_run_pipeline(),
)
def test_destroy_snapshot(mocked_subprocess_run):
    zsm.zfs.destroy_snapshot(
        snapshot=zsm.zfs.Snapshot(dataset=zsm.zfs.Dataset(name="tank/a"), name="asdf")
    )
    mocked_subprocess_run.assert_called()



# SPDX-License-Identifier: BSD-2-Clause
from unittest import mock

import pid
import pkg_resources
from click.testing import CliRunner

import zsm
import zsm.config


@mock.patch("zsm.cli.open", mock.mock_open())
@mock.patch("zsm.config.load")
def test_load_config_calls_config_load(mock_config_load):
    zsm.cli.load_config()

    mock_config_load.assert_called()


@mock.patch("pid.PidFile", mock.MagicMock())
@mock.patch("zsm.cli.load_config", mock.MagicMock(return_value={}))
@mock.patch("zsm.manage.manage_snapshots", mock.MagicMock())
def test_cron_returns_0_on_success():
    runner = CliRunner()
    result = runner.invoke(
        zsm.cli.program,
        ["cron", "--log-file", "/dev/null", "--config-file", "/dev/null"],
        obj={},
    )

    assert result.exit_code == 0


@mock.patch("pid.PidFile", mock.MagicMock())
@mock.patch("zsm.cli.load_config", mock.MagicMock(return_value=None))
@mock.patch("zsm.manage.manage_snapshots", mock.MagicMock())
def test_cron_returns_1_on_failed_config_load():
    runner = CliRunner()
    result = runner.invoke(zsm.cli.program, ["cron", "--log-file", "/dev/null"], obj={})

    assert result.exit_code == 1


@mock.patch("pid.PidFile", mock.MagicMock(side_effect=pid.PidFileAlreadyLockedError))
@mock.patch("zsm.cli.load_config", mock.MagicMock(return_value={}))
@mock.patch("zsm.manage.manage_snapshots", mock.MagicMock())
def test_cron_returns_2_on_pidfilealreadylockederror():
    runner = CliRunner()
    result = runner.invoke(zsm.cli.program, ["cron", "--log-file", "/dev/null"], obj={})

    assert result.exit_code == 2


@mock.patch("zsm.cli.open", mock.mock_open())
@mock.patch("zsm.config.load", mock.MagicMock())
def test_validate_config_returns_0_on_success():
    runner = CliRunner()
    result = runner.invoke(
        zsm.cli.program, ["validate-config", "--config-file", "/fake/path"], obj={}
    )

    assert result.exit_code == 0


@mock.patch("zsm.cli.open", mock.mock_open())
@mock.patch("zsm.config.load", mock.MagicMock(side_effect=zsm.config.ValidationError))
def test_validate_config_returns_1_on_validationerror():
    runner = CliRunner()
    result = runner.invoke(zsm.cli.program, ["validate-config"], obj={})

    assert result.exit_code == 1


@mock.patch("zsm.cli.open", mock.MagicMock(side_effect=FileNotFoundError))
def test_validate_config_returns_1_on_filenotfounderror():
    runner = CliRunner()
    result = runner.invoke(zsm.cli.program, ["validate-config"], obj={})

    assert result.exit_code == 1


def test_version_prints_version():
    runner = CliRunner()
    result = runner.invoke(zsm.cli.program, ["version"], obj={})

    assert result.exit_code == 0

    version = pkg_resources.get_distribution("zsm").version
    assert result.output.strip() == f"zsm {version}"


@mock.patch("zsm.cli.program")
def test_cli_calls_program(mock_program):
    zsm.cli.cli()
    mock_program.assert_called()



# SPDX-License-Identifier: BSD-2-Clause
import datetime
from unittest import mock

import zsm.manage
import zsm.zfs


def test_get_snapshots_finds_correct_items():
    label = "abc"
    valid_names = ["zsm_abc_11", "zsm_abc_22"]
    invalid_names = [
        "zsm_xyz_11",  # Wrong label.
        "zsm_abcc_11",  # Label is only partial match.
        "xyz_abc_11",  # Right pattern, but wrong prefix.
        "abc",  # Wrong prefix, and not enough parts.
        "zsm",  # Right prefix, but not enough parts.
        "zsm_abc_11_lol",  # Right prefix, and right label, but too many parts.
        "",  # Empty string should not even be possible to get.
    ]
    names = valid_names + invalid_names

    dataset = mock.MagicMock()

    with mock.patch(
        "zsm.zfs.get_snapshots",
        return_value=[zsm.zfs.Snapshot(dataset=dataset, name=name) for name in names],
    ):
        snapshots = zsm.manage.get_snapshots(dataset=dataset, label=label)

    # Assert that result only contains snapshots with valid names.
    assert snapshots == [
        zsm.zfs.Snapshot(dataset=dataset, name=name) for name in valid_names
    ]


def test_create_snaphot_calls_zfs_create_snapshot():
    dataset = mock.MagicMock()

    with mock.patch("zsm.zfs.create_snapshot") as mock_create_snapshot:
        zsm.manage.create_snapshot(dataset=dataset, label="asdf")

    assert mock_create_snapshot.called


def test_parse_snapshot_returns_correct_parts():
    expected_label = "daily"
    expected_timestamp = datetime.datetime.now()
    formatted_expected_timestamp = expected_timestamp.strftime(
        zsm.manage.SNAPSHOT_TIMESTAMP_FORMAT
    )

    label, timestamp = zsm.manage.parse_snapshot(
        snapshot=zsm.zfs.Snapshot(
            dataset=mock.MagicMock(),
            name=zsm.manage.SNAPSHOT_DELIMITER.join(
                [
                    zsm.manage.SNAPSHOT_PREFIX,
                    expected_label,
                    formatted_expected_timestamp,
                ]
            ),
        )
    )

    # Round down microseconds, because that is not stored in the snapshot name.
    expected_timestamp = expected_timestamp.replace(microsecond=0)

    assert label == expected_label
    assert timestamp == expected_timestamp



# SPDX-License-Identifier: BSD-2-Clause
import setuptools


setuptools.setup(
    name="zsm",
    version="0.3.0",
    keywords="zfs snapshots freebsd linux",
    description="ZFS Snapshot Manager",
    long_description="Please see the project links.",
    project_urls={
        "Documentation": "https://zsm.readthedocs.io/",
        "Source": "https://github.com/thnee/zsm",
    },
    license="BSD-2-Clause",
    author="Mattias Lindvall",
    author_email="mattias.lindvall@gmail.com",
    package_dir={"": "src"},
    packages=["zsm"],
    python_requires=">=3.7",
    install_requires=[
        "click~=8.0.1",
        "pid~=3.0.4",
        "pyyaml~=5.4.1",
        "marshmallow~=3.13.0",
    ],
    extras_require={
        "test": [
            "pytest~=6.2.5",
            "pytest-cov~=2.12.1",
            "flake8~=3.9.2",
            "flake8-bugbear~=21.4.3",
            "black~=21.8b0",
            "isort~=5.9.3",
        ],
        "build": [
            "wheel~=0.37.0",
            "twine~=3.4.2",
        ],
    },
    entry_points={"console_scripts": ["zsm = zsm.cli:cli"]},
    classifiers=[
        # "Development Status :: 1 - Planning",
        # "Development Status :: 2 - Pre-Alpha",
        # "Development Status :: 3 - Alpha",
        "Development Status :: 4 - Beta",
        # "Development Status :: 5 - Production/Stable",
        # "Development Status :: 6 - Mature",
        # "Development Status :: 7 - Inactive",
        "Environment :: Console",
        "Intended Audience :: System Administrators",
        "Operating System :: POSIX :: BSD :: FreeBSD",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: System :: Filesystems",
    ],
)



# SPDX-License-Identifier: BSD-2-Clause
import datetime
import platform
import re
from pathlib import Path

import marshmallow
import yaml
from marshmallow import Schema, fields

from . import zfs


class ValidationError(Exception):
    pass


class DatasetField(fields.String):
    def _deserialize(self, value, attr, obj, **kwargs):
        return zfs.Dataset(name=value)

    def _validate(self, value):
        datasets = zfs.get_datasets()

        for dataset in datasets:
            if dataset == value:
                return dataset

        raise marshmallow.ValidationError("Dataset does not exist")


class LabelField(fields.String):
    pattern = r"[^a-zA-Z0-9-]"

    def _validate(self, value):
        matches = re.findall(self.pattern, value)
        if matches:
            formatted_matches = ", ".join([f"'{m}'" for m in matches])
            raise marshmallow.ValidationError(
                f"Contains invalid characters: {formatted_matches}"
            )

        return value


class FrequencyField(fields.String):
    pattern = r"(?P<value>\d+)(?P<unit>[a-zA-Z]+)"

    units = {
        "seconds": ["s", "sec", "second", "secs", "seconds"],
        "minutes": ["m", "min", "minute", "minutes"],
        "hours": ["h", "hour", "hours"],
        "days": ["d", "day", "days"],
        "weeks": ["w", "week", "weeks"],
    }

    def _deserialize(self, value, attr, obj, **kwargs):
        result = {}

        pattern = re.compile(self.pattern)
        matches = pattern.finditer(value)
        for match in matches:
            group = match.groupdict()

            found_unit = None

            for unit, aliases in self.units.items():
                if group["unit"].lower() in aliases:
                    found_unit = unit

            if found_unit is not None:
                result[found_unit] = int(group["value"])
            else:
                raise marshmallow.ValidationError(
                    f"Unit not supported: {group['unit']}"
                )

        return datetime.timedelta(**result)


class SnapshotSchema(Schema):
    dataset = DatasetField(required=True)
    label = LabelField(required=True)
    frequency = FrequencyField(required=True)
    retention = fields.Integer(required=True)


class ConfigSchema(Schema):
    snapshots = fields.Nested(SnapshotSchema, many=True, required=True)


def get_platform_path() -> Path:
    paths = {"FreeBSD": "/usr/local/etc/", "Linux": "/etc/"}
    system = platform.system()
    return Path(paths[system])


def get_path() -> Path:
    return get_platform_path() / "zsm.yaml"


def load(data: str) -> dict:
    try:
        data = yaml.safe_load(data)

    except yaml.YAMLError as e:
        msg = "Invalid YAML"

        problem = getattr(e, "problem", None)
        if problem is not None:
            msg += f": {e.problem}"

        problem_mark = getattr(e, "problem_mark", None)
        if problem_mark is not None:
            msg += (
                f": line={e.problem_mark.line + 1} column={e.problem_mark.column + 1}"
            )

        raise ValidationError(msg)

    try:
        data = ConfigSchema().load(data)

    except marshmallow.ValidationError as e:
        raise ValidationError(e.messages)

    return data



# SPDX-License-Identifier: BSD-2-Clause
import datetime
import logging
from typing import Dict, Iterable, Tuple

from . import zfs


log = logging.getLogger(__name__)

SNAPSHOT_DELIMITER = "_"
SNAPSHOT_PREFIX = "zsm"
SNAPSHOT_TIMESTAMP_FORMAT = "%Y-%m-%dT%H:%M:%S"


def get_snapshots(dataset: zfs.Dataset, label: str) -> Iterable[zfs.Snapshot]:
    snapshots = zfs.get_snapshots(dataset=dataset)

    result = []

    for snapshot in snapshots:
        parts = snapshot.name.split(SNAPSHOT_DELIMITER)

        if len(parts) != 3:
            continue

        if parts[0] == SNAPSHOT_PREFIX and parts[1] == label:
            result.append(snapshot)

    return result


def create_snapshot(dataset: zfs.Dataset, label: str) -> None:
    timestamp = datetime.datetime.now().strftime(SNAPSHOT_TIMESTAMP_FORMAT)
    zfs.create_snapshot(
        dataset=dataset,
        name=SNAPSHOT_DELIMITER.join([SNAPSHOT_PREFIX, label, timestamp]),
    )


def parse_snapshot(snapshot: zfs.Snapshot) -> Tuple[str, datetime.datetime]:
    _, label, timestamp = snapshot.name.split(SNAPSHOT_DELIMITER)
    timestamp = datetime.datetime.strptime(timestamp, SNAPSHOT_TIMESTAMP_FORMAT)
    return label, timestamp


def manage_snapshots(config: Dict, now: datetime.datetime, dry_run: bool) -> None:
    for snapshot_config in config["snapshots"]:
        create_snapshots(
            now=now,
            dataset=snapshot_config["dataset"],
            label=snapshot_config["label"],
            frequency=snapshot_config["frequency"],
            dry_run=dry_run,
        )

    for snapshot_config in config["snapshots"]:
        destroy_snapshots(
            dataset=snapshot_config["dataset"],
            label=snapshot_config["label"],
            retention=snapshot_config["retention"],
            dry_run=dry_run,
        )


def create_snapshots(
    now: datetime.datetime,
    dataset: zfs.Dataset,
    label: str,
    frequency: datetime.timedelta,
    dry_run: bool,
) -> None:
    snapshots = get_snapshots(dataset=dataset, label=label)

    if len(snapshots) == 0:
        log.info(f"[{dataset.name}:{label}] No snapshots yet, creating the first one.")

        if not dry_run:
            create_snapshot(dataset=dataset, label=label)

    else:
        latest_snapshot = snapshots[0]
        _, latest_timestamp = parse_snapshot(snapshot=latest_snapshot)
        latest_age = now - latest_timestamp

        if latest_age > frequency:
            log.info(
                f"[{dataset.name}:{label}] "
                f"Latest snapshot ({latest_snapshot.name}) is {latest_age} old, "
                "creating new."
            )

            if not dry_run:
                create_snapshot(dataset=dataset, label=label)

        else:
            log.info(
                f"[{dataset.name}:{label}] "
                f"Latest snapshot ({latest_snapshot.name}) is only {latest_age} old, "
                "skipping."
            )


def destroy_snapshots(
    dataset: zfs.Dataset, label: str, retention: int, dry_run: bool
) -> None:
    any_old_found = False

    while True:
        snapshots = get_snapshots(dataset=dataset, label=label)

        if len(snapshots) > retention:
            oldest_snapshot = snapshots[-1]

            log.info(
                f"[{dataset.name}:{label}] "
                f"Found old snapshot ({oldest_snapshot.name}), destroying it."
            )

            if not dry_run:
                zfs.destroy_snapshot(snapshot=oldest_snapshot)

            any_old_found = True

        else:
            break

    if not any_old_found:
        log.info(f"[{dataset.name}:{label}] There are no old snapshots to destroy.")
